#!/usr/bin/perl

use v5.10;
use strict;
use warnings;
binmode STDOUT, ":utf8";

use XML::Parser;
use Date::Calc qw(Delta_Days);
use HTML::Entities;
use Getopt::Std;
#use Data::Dumper;

my $version = "1.7";

my %fields = ();
my @MCepoch = (1899, 12, 31);

# Map XML path to MC fields and the MC data type.  For character data only.
my %path_to_MC_field = (
    'Collection::DVD::Audio::AudioTrack::AudioContent'	=> [ 'Language',		\&add_list_item ],
    'Collection::DVD::Audio::AudioTrack::AudioChannels'	=> [ 'Language',		\&add_audiotrack ],
    'Collection::DVD::Audio::AudioTrack::AudioFormat'	=> [ 'Language',		\&add_audiotrack ],

    'Collection::DVD::CountryOfOrigin'			=> [ 'Country',			\&add_list_item ],
    'Collection::DVD::Discs::Disc::Location'		=> [ 'Filename',		sub { $_[1] and $fields{$_[0]}[-1] .= $_[1] } ],
    'Collection::DVD::Format::FormatAspectRatio'	=> [ 'Aspect Ratio',		\&add_single_item ],
    'Collection::DVD::Format::FormatVideoStandard'	=> [ 'Video Standard',		\&add_single_item ],
    'Collection::DVD::Genres::Genre'			=> [ 'Genre', 			\&add_list_item ],
    'Collection::DVD::ID'				=> [ 'Image File',		sub { $fields{$_[0]} = $_[1] . 'f.jpg' } ],
    'Collection::DVD::OriginalTitle'			=> [ 'Original Title',		\&add_single_item ],
    'Collection::DVD::Overview'				=> [ 'Description',		\&add_single_item ],
    'Collection::DVD::ProductionYear'			=> [ 'Date',			sub { $_[1] and $fields{$_[0]} = 1 + Delta_Days(@MCepoch, ($_[1],1,1)) } ],
    'Collection::DVD::Rating'				=> [ 'MPAA Rating',		\&add_single_item ],
    'Collection::DVD::RatingSystem'			=> [ 'MPAA Rating Description',	\&add_single_item ],
    'Collection::DVD::RunningTime'			=> [ 'Duration',		\&add_single_item ],
    'Collection::DVD::Studios::Studio'			=> [ 'Studios',			\&add_list_item ],
    'Collection::DVD::Title'				=> [ 'Name',			\&add_single_item ],
    'Collection::DVD::UPC'				=> [ 'UPC',			\&add_single_item ],

    # custom user fields
    'Collection::DVD::Subtitles::Subtitle'		=> [ 'Subtitles',		\&add_list_item ],
    'Collection::DVD::CollectionNumber'			=> [ 'Collection Number',	\&add_single_item ],
    'Collection::DVD::Events::Event::EventType'		=> [ 'Watched',			sub { $_[1] eq 'Watched' and $fields{$_[0]} = 'Yes' } ],
    'Collection::DVD::MediaTypes::DVD'			=> [ 'Media Support',		\&add_mediatypes ],
    'Collection::DVD::MediaTypes::HDDVD'		=> [ 'Media Support',		\&add_mediatypes ],
    'Collection::DVD::MediaTypes::BluRay'		=> [ 'Media Support',		\&add_mediatypes ],
    'Collection::DVD::MediaTypes::CustomMediaType'	=> [ 'Media Support',		\&add_mediatypes ],
);

# -a: all entries (include those with no Filename)
our ($opt_a);
getopts('a');

my $file = shift;

die "Can't find file \"$file\""
    unless -f $file;

my $parser = new XML::Parser(ErrorContext => 2);

$parser->setHandlers(
    Char =>	\&char_handler,
    Start =>	\&start_handler,
    End =>	\&end_handler,
    Final =>	\&final_handler,
    Default =>	\&default_handler
);

sub print_mpl;

print_mpl '<?xml version="1.0" encoding="UTF-8" standalone="yes" ?>';
print_mpl '<MPL Version="2.0" Title="JRSidecar">';

$parser->parsefile($file);

print_mpl '</MPL>';

# end


# handlers below

my @paths;		# xml paths

sub start_handler {
    my ($p, $el) = @_;

    push @paths, my $path = join('::', $p->context, $el);

    sub get_values {
	my $attr_re = shift;
	my $ret = '';

	while (@_) {
	    my ($a, $v) = (shift, shift);
	    $ret = ($ret ? join($1 ? '=' : ' ', $ret, $v) : $v)  if $v and $a =~ /$attr_re/;
	}
	return $ret;
    }

    if (defined $p->current_element) {

	if ($path eq 'Collection::DVD') {
	    %fields = ();
	}

	elsif ($path eq 'Collection::DVD::Discs::Disc') {
	    #say "START DISC: ";
	    push @{$fields{'Filename'}}, '';
	}

	elsif ($path eq 'Collection::DVD::Actors::Actor') {
	    push @{$fields{'Actors'}}, get_values('^(?:First|Middle|Last)Name|(Role)', @_);
	}

	elsif ($path eq 'Collection::DVD::Credits::Credit') {
	    my $val = get_values('^(?:(?:First|Middle|Last)Name)|(CreditSubtype)$', @_);
	    my $key = ($val =~ s/=(Composer|Director|Screenwriter)$//) ? $1 : 'Credits';
	    add_list_item_unique($key, $val);
	    #add_list_item_unique('Artist', $val)	 if $key eq 'Director';	# dual mapped: not desired by sonata31
	    add_list_item_unique('Music By', $val)	 if $key eq 'Composer';	# dual mapped
	}

	elsif ($path eq 'Collection::DVD::Tags::Tag') {
	    my $val = get_values('^Name$', @_);
	    #<Tag Name="Season 1" FullName="Season 1"/>
	    #<Tag Name="Series:Alias" FullName="Series:Alias"/>
	    if ($val =~ /^(Season) (\d+)$/ or
		$val =~ /^(Series):(.+)$/) {
		add_single_item($1, $2);
		$fields{'Media Sub Type'}	= 'TV Show';
	    }
	}
    }
}

sub char_handler {
    my ($p, $data) = @_;

    my $path = $paths[-1];

    # the expat parser returns entities as single characters
    if ($data eq '&' or $data eq '<' or $data eq '>') { $data = encode_entities($data); }
    else					      { $data = decode_entities($data); }

    # call the field mapping hander
    my ($mcfield,$coderef);
    if (exists $path_to_MC_field{$path} and my $ref = $path_to_MC_field{$path}) {
	# coderef  (mc field,      data,  path)
	$ref->[1]($ref->[0], $data, $path);
    }
}

sub end_handler {
    my ($p, $el) = @_;

    my $path = pop @paths;

    if ($path eq 'Collection::DVD') {

	$fields{'Media Type'}	  = 'Video' if !exists $fields{'Media Type'};
	$fields{'Media Sub Type'} = 'Movie' if !exists $fields{'Media Sub Type'};

	if ($fields{'Filename'}[0]) {
	    for (@{$fields{'Filename'}}) {
		if (/\\([^\\]+?)\s*-\s*S(\d+)E(\d+)\s*-\s*/) {
		    #say "Filename: Series: $1, Season $2, Episode $3";
		    $fields{'Series'} = $1 if $1 ne '';
		    $fields{'Season'} = $2 if $2 ne '';
		    $fields{'Episode'} = $3 if $3 ne '';
		    $fields{'Media Sub Type'}	= 'TV Show';
		}

		print_mpl '<Item>';

		for my $key (sort keys %fields) {
		    print_mpl '<Field Name="', $key, '">',
			($key eq 'Filename' ? $_ : 
			      ref $fields{$key} eq 'ARRAY' ? join('; ', @{$fields{$key}}) : $fields{$key}),
		    '</Field>';

		}

		print_mpl '</Item>';
	    }
	}
	elsif ($opt_a) {
		print_mpl '<Item>';
		for my $key (sort keys %fields) {
		    print_mpl '<Field Name="', $key, '">',
				  ref $fields{$key} eq 'ARRAY' ? join('; ', @{$fields{$key}}) : $fields{$key},
			      '</Field>';

		}

		print_mpl '</Item>';
	}
    }

    elsif ($path eq 'Collection::DVD::Discs::Disc') {
	#say "END DISC";
	if (exists $fields{'Filename'}[-1]) {
	    if ($fields{'Filename'}[-1] eq '') {
		# remove non-file disc placeholders
		pop @{$fields{'Filename'}};
	    }
	    else {
		# sometimes filenames are double-quoted in DVD Profiler.
		$fields{'Filename'}[-1] =~ s/"//g
	    }
	}
    }
}

sub final_handler {
    #print Dumper(\%fields);
}

sub default_handler {
}

sub print_mpl {
    local $\ = "\r\n";

    print @_;

}

sub add_list_item_unique
{
    push @{$fields{$_[0]}}, $_[1]  if ! exists $fields{$_[0]} or !grep(/^\Q$_[1]\E$/, @{$fields{$_[0]}});
}

sub add_list_item
{
    push @{$fields{$_[0]}}, $_[1];
}

sub add_single_item
{
    $fields{$_[0]} .= $_[1];
}

sub add_audio_channels {

    my %ch_map = ( 
	'Mono'		     => 1,
	'2-Channel Stereo'   => 2,
	'3.1'		     => 4,
	'Dolby Surround'     => 5,
	'5.0'		     => 5,
	'5.1'		     => 6,
	'5.1 (Matrixed 6.1)' => 7,
	'6.1 (Discrete)'     => 7,
	'7.1'		     => 8,
    );

    $fields{$_[0]} = $ch_map{$_[1]}  if !exists($fields{$_[0]}) and exists($ch_map{$_[1]});
}

sub add_audiotrack {
    #<AudioContent>English</AudioContent>
    #<AudioFormat>Dolby Digital</AudioFormat>
    #<AudioChannels>5.1</AudioChannels>

    # this relies on the order content,format,channels in the XML
    my $key = (split('::', $_[2]))[-1];
    if ($key eq 'AudioFormat') {
	$fields{$_[0]}->[-1] .= " ($_[1]";
    }
    elsif ($key eq 'AudioChannels') {
	add_audio_channels('Channels', $_[1], $_[2]);
	$fields{$_[0]}->[-1] .= " $_[1])";
    }

}

sub add_mediatypes {
    my $key = (split('::', $_[2]))[-1];
    if ($_[1] eq 'true') {
	$fields{$_[0]} = exists $fields{$_[0]} ? join('; ', $fields{$_[0]}, $key) : $key;
    }
    elsif ($key eq 'CustomMediaType') {
	$fields{$_[0]} = exists $fields{$_[0]} ? join('; ', $fields{$_[0]}, $_[1]) : $_[1];
    }
}

